let string_sz = 17
let keywords_sz = 17

open Util

open Parser

open Big_int

type info = {
  filename: string;
  line: int;
  bol: int;
}

let update lexbuf lexbuf' =
  lexbuf'.Lexing.lex_start_p <- {
    lexbuf'.Lexing.lex_curr_p with
      Lexing.pos_cnum = Ulexing.lexeme_start lexbuf
  };
  lexbuf'.Lexing.lex_curr_p <- {
    lexbuf'.Lexing.lex_curr_p with
      Lexing.pos_cnum = Ulexing.lexeme_end lexbuf
  }

let keywords = Utf8.Hashtbl.create keywords_sz

let add_keyword (ident, keyword) =
  Utf8.Hashtbl.add keywords ident keyword

let () =
  List.iter add_keyword
    [
      (Utf8.of_latin1 "_", UNDERSCORE);
      (Utf8.of_latin1 "and", AND);
      (Utf8.of_latin1 "call", CALL);
      (Utf8.of_latin1 "do", DO);
      (Utf8.of_latin1 "else", ELSE);
      (Utf8.of_latin1 "exists", EXISTS);
      (Utf8.of_latin1 "for", FOR);
      (Utf8.of_latin1 "forall", FORALL);
      (Utf8.of_latin1 "function", FUNCTION);
      (Utf8.of_latin1 "if", IF);
      (Utf8.of_latin1 "in", IN);
      (Utf8.of_latin1 "let", LET);
      (Utf8.of_latin1 "mod", MOD);
      (Utf8.of_latin1 "module", MODULE);
      (Utf8.of_latin1 "namespace", NAMESPACE);
      (Utf8.of_latin1 "not", NOT);
      (Utf8.of_latin1 "or", OR);
      (Utf8.of_latin1 "package", PACKAGE);
      (Utf8.of_latin1 "this", THIS);
      (Utf8.of_latin1 "with", WITH)
    ]

let read_int v =
  let l = String.length v in
  let rec read_hex_int accu i =
    if i >= l then accu
    else
      let c = v.[i] in
      let u =
        match c with
          | 'a' .. 'f' -> int_of_char c - int_of_char 'a' + 0xA
          | 'A' .. 'F' -> int_of_char c - int_of_char 'A' + 0xA
          | '0' .. '9' -> int_of_char c - int_of_char '0'
          | _ -> failwith "read_int" in
      let accu' = add_int_big_int u (mult_int_big_int 0x10 accu) in
      read_hex_int accu' (succ i) in
  let rec read_oct_int accu i =
    if i >= l then accu
    else
      let c = v.[i] in
      let u =
        match c with
          | '0' .. '7' -> int_of_char c - int_of_char '0'
          | _ -> failwith "read_int" in
      let accu' = add_int_big_int u (mult_int_big_int 0o10 accu) in
      read_hex_int accu' (succ i) in
  let rec read_bin_int accu i =
    if i >= l then accu
    else
      let c = v.[i] in
      let u =
        match c with
          | '0' .. '1' -> int_of_char c - int_of_char '0'
          | _ -> failwith "read_int" in
      let accu' = add_int_big_int u (mult_int_big_int 0b10 accu) in
      read_hex_int accu' (succ i) in
  try
    let i = String.index v 'e' in
    let m = big_int_of_string (String.sub v 0 i) in
    let e =
      big_int_of_string (String.sub v (i + 1) (l - i - 1)) in
    mult_big_int m (power_int_positive_big_int 10 e)
  with Not_found ->
    if l >= 2 && v.[0] = '0' then
      match v.[1] with
        | 'x' -> read_hex_int zero_big_int 2
        | 'o' -> read_oct_int zero_big_int 2
        | 'b' -> read_bin_int zero_big_int 2
        | _ -> big_int_of_string v
    else
      big_int_of_string v

let regexp letter = xml_letter

let regexp digit = ['0' - '9']

let regexp int = digit* | "0x" (digit | ['A' - 'F' 'a' - 'f'])*

let regexp float =
  (['0' - '9']+ "." ['0' - '9']* ("e" ['+' '-']? ['0' - '9']+)?
  | ['0' - '9']+ ("e-" ['0' - '9']+)?)

let regexp ident_start = letter | '_'

let regexp ident_body = ident_start | digit | ['\'' '-']

let regexp ident = ident_start ident_body*

let regexp blank = [' ' '\t' '\r']

let escaped_char lexbuf' = lexer
  | 'n' -> '\n'
  | 'r' -> '\r'
  | 't' -> '\t'
  | 'b' -> '\b'
  | '\'' -> '\''
  | '\"' -> '\"'
  | '\\' -> '\\'
  | _ ->
      update lexbuf lexbuf';
      Err.unknown_escaped_char (Pos.of_lexbuf lexbuf')

let rec simple_quoted_string buffer lexbuf' = lexer
  | '\'' ->
      update lexbuf lexbuf';
      let s = Buffer.contents buffer in
      ATOM s
  | "\'(" ->
      update lexbuf lexbuf';
      FUNCTOR (Buffer.contents buffer)
  | [^ '\\' '\''] ->
      Buffer.add_string buffer (Ulexing.utf8_lexeme lexbuf);
      simple_quoted_string buffer lexbuf' lexbuf
  | '\\' ->
      let c = escaped_char lexbuf' lexbuf in
      Buffer.add_char buffer c;
      simple_quoted_string buffer lexbuf' lexbuf
  | eof | '\n' ->
      update lexbuf lexbuf';
      Err.unterminated_string (Pos.of_lexbuf lexbuf')

let rec double_quoted_string buffer lexbuf' = lexer
  | '"' ->
      update lexbuf lexbuf';
      Buffer.contents buffer
  | [^ '\\' '"'] ->
      Buffer.add_string buffer (Ulexing.utf8_lexeme lexbuf);
      double_quoted_string buffer lexbuf' lexbuf
  | '\\' ->
      let c = escaped_char lexbuf' lexbuf in
      Buffer.add_char buffer c;
      double_quoted_string buffer lexbuf' lexbuf
  | eof | '\n' ->
      update lexbuf lexbuf';
      Err.unterminated_string (Pos.of_lexbuf lexbuf')

let incr_line lexbuf' =
  lexbuf'.Lexing.lex_curr_p <- {
    lexbuf'.Lexing.lex_curr_p with
      Lexing.pos_lnum =
      succ (lexbuf'.Lexing.lex_curr_p.Lexing.pos_lnum);
      Lexing.pos_bol =
      lexbuf'.Lexing.lex_curr_p.Lexing.pos_cnum;
  }

let rec line_comment lexbuf' = lexer
  | '\n' | eof ->
      update lexbuf lexbuf';
      incr_line lexbuf';
      ()
  | [^ '\n' ] ->
      line_comment lexbuf' lexbuf

let rec long_comment lexbuf' = lexer
  | "*/" -> ()
  | '\n' ->
      update lexbuf lexbuf';
      incr_line lexbuf';
      long_comment lexbuf' lexbuf
  | [^ '\n' ] ->
      long_comment lexbuf' lexbuf
  | eof ->
      update lexbuf lexbuf';
      Err.unterminated_comment (Pos.of_lexbuf lexbuf')

let rec code_comment lexbuf' = lexer
  | "*)" -> ()
  | "(*" ->
      code_comment lexbuf' lexbuf;
      code_comment lexbuf' lexbuf
  | "/*" ->
      update lexbuf lexbuf';
      long_comment lexbuf' lexbuf;
      code_comment lexbuf' lexbuf
  | "//" ->
      update lexbuf lexbuf';
      line_comment lexbuf' lexbuf;
      code_comment lexbuf' lexbuf
  | '\n' ->
      update lexbuf lexbuf';
      incr_line lexbuf';
      code_comment lexbuf' lexbuf
  | '"' ->
      update lexbuf lexbuf';
      ignore (double_quoted_string (Buffer.create string_sz) lexbuf' lexbuf);
      code_comment lexbuf' lexbuf
  | '\'' ->
      update lexbuf lexbuf';
      ignore (simple_quoted_string (Buffer.create string_sz) lexbuf' lexbuf);
      code_comment lexbuf' lexbuf
  | [^ '\n' '"' '\'' ] ->
      code_comment lexbuf' lexbuf
  | eof ->
      update lexbuf lexbuf';
      Err.unterminated_comment (Pos.of_lexbuf lexbuf')

let push stack lexbuf' symbol_start item =
  let hd = (symbol_start, Pos.of_lexbuf lexbuf', item) in
  stack := hd :: !stack

let pop stack lexbuf' symbol_start symbol_end =
  match !stack with
    | [] -> Err.unmatched_delimiter (Pos.of_lexbuf lexbuf') symbol_end
    | (symbol_start', pos_start, _) :: tail ->
        if symbol_start <> symbol_start' then
          begin
            let pos = {
              Pos.first = pos_start.Pos.first;
              last = Pos.of_lexing lexbuf'.Lexing.lex_curr_p;
            } in
            Err.mismatched_delimiter pos symbol_start' symbol_end
          end;
        stack := tail

type lex =
    Lex of
      ((string * Pos.range * lex) list ref
       -> Lexing.lexbuf -> Ulexing.lexbuf -> token)

let lex_ident stack lexbuf' lexbuf =
  update lexbuf lexbuf';
  let lexeme = Ulexing.utf8_lexeme lexbuf in
  begin
    try Utf8.Hashtbl.find keywords lexeme
    with Not_found ->
      IDENT lexeme
  end

let rec main stack lexbuf' = lexer
  | blank ->
      main stack lexbuf' lexbuf
  | 0x2200 -> FORALL (* ∀ *)
  | 0x2203 -> EXISTS (* ∃ *)
  | 0x2192 -> ARROW (* → *)
  | 0x21D2 -> DBL_ARROW (* ⇒ *)
  | 0x2190 -> LEFT_ARROW (* ← *)
  | 0x2254 -> COLON_EQ (* ≔ *)
  | 0x2260 -> NEQ (* ≠ *)
  | 0x2264 -> LE (* ≤ *)
  | 0x2265 -> GE (* ≥ *)
  | '\n' ->
      update lexbuf lexbuf';
      incr_line lexbuf';
      main stack lexbuf' lexbuf
  | "(*" ->
      update lexbuf lexbuf';
      code_comment lexbuf' lexbuf;
      main stack lexbuf' lexbuf
  | "/*" ->
      update lexbuf lexbuf';
      long_comment lexbuf' lexbuf;
      main stack lexbuf' lexbuf
  | "//" ->
      update lexbuf lexbuf';
      line_comment lexbuf' lexbuf;
      main stack lexbuf' lexbuf
  | ident -> lex_ident stack lexbuf' lexbuf
  | ident "(" ->
      update lexbuf lexbuf';
      let lexeme_length = Ulexing.lexeme_length lexbuf - 1 in
      let lexeme = Ulexing.utf8_sub_lexeme lexbuf 0 lexeme_length in
      push stack lexbuf' "(" (Lex main);
      IDENT_PAR lexeme
  | ident "[" ->
      update lexbuf lexbuf';
      let lexeme_length = Ulexing.lexeme_length lexbuf - 1 in
      let lexeme = Ulexing.utf8_sub_lexeme lexbuf 0 lexeme_length in
      push stack lexbuf' "[" (Lex main);
      IDENT_SQ lexeme
  | '@' int ->
      update lexbuf lexbuf';
      let lexeme_length = Ulexing.lexeme_length lexbuf - 1 in
      let lexeme = Ulexing.utf8_sub_lexeme lexbuf 1 lexeme_length in
      KERNEL_CONST (Kernel.Const.Int (read_int lexeme))
  | '@' float ->
      update lexbuf lexbuf';
      let lexeme_length = Ulexing.lexeme_length lexbuf - 1 in
      let lexeme = Ulexing.utf8_sub_lexeme lexbuf 1 lexeme_length in
      KERNEL_CONST (Kernel.Const.Float (float_of_string lexeme))
  | "@\"" ->
      update lexbuf lexbuf';
      let s = double_quoted_string (Buffer.create string_sz) lexbuf' lexbuf in
      KERNEL_CONST (Kernel.Const.String s)
  | int ->
      update lexbuf lexbuf';
      let lexeme = Ulexing.utf8_lexeme lexbuf in
      INT (read_int lexeme)
  | float ->
      update lexbuf lexbuf';
      let lexeme = Ulexing.utf8_lexeme lexbuf in
      FLOAT (float_of_string lexeme)
(*  | digit+ "." digit* ->
      update lexbuf lexbuf';
      let lexeme = Ulexing.utf8_lexeme lexbuf in
      FLOAT (float_of_string lexeme) *)
  | '\'' ->
      update lexbuf lexbuf';
      let lexem =
        simple_quoted_string (Buffer.create string_sz) lexbuf' lexbuf in
      begin
        match lexem with
          | FUNCTOR _ ->
                push stack lexbuf' "(" (Lex main)
          | _ -> ()
      end;
      lexem
  | '"' ->
      update lexbuf lexbuf';
      STRING (double_quoted_string (Buffer.create string_sz) lexbuf' lexbuf)
  | "(" -> lex_par_open stack lexbuf' lexbuf
  | ")" ->
      update lexbuf lexbuf';
      pop stack lexbuf' "(" ")";
      PAR_CLOSE
  | ")[" ->
      update lexbuf lexbuf';
      pop stack lexbuf' "(" ")";
      push stack lexbuf' "[" (Lex main);
      PAR_SQ
  | "[" ->
      update lexbuf lexbuf';
      push stack lexbuf' "[" (Lex main);
      SQ_OPEN
  | "]" ->
      update lexbuf lexbuf';
      pop stack lexbuf' "[" "]";
      SQ_CLOSE
  | "][" ->
      update lexbuf lexbuf';
      pop stack lexbuf' "]" "[";
      push stack lexbuf' "[" (Lex main);
      SQ_SQ
  | "{" -> lex_curly_open stack lexbuf' lexbuf
  | "}" ->
      update lexbuf lexbuf';
      pop stack lexbuf' "{" "}";
      CURLY_CLOSE
  | "," ->
      update lexbuf lexbuf';
      COMMA
  | "." ->
      update lexbuf lexbuf';
      DOT
  | ".." ->
      update lexbuf lexbuf';
      DOTDOT
  | "..." ->
      update lexbuf lexbuf';
      DOTDOTDOT
  | ":" ->
      update lexbuf lexbuf';
      COLON
  | ";" ->
      update lexbuf lexbuf';
      SEMICOLON
  | "=" ->
      update lexbuf lexbuf';
      EQ
  | "<>" ->
      update lexbuf lexbuf';
      NEQ
  | "->" ->
      update lexbuf lexbuf';
      ARROW
  | "=>" ->
      update lexbuf lexbuf';
      DBL_ARROW
  | "<-" ->
      update lexbuf lexbuf';
      LEFT_ARROW
  | ":=" ->
      update lexbuf lexbuf';
      COLON_EQ
  | "+" ->
      update lexbuf lexbuf';
      PLUS
  | "-" ->
      update lexbuf lexbuf';
      MINUS
  | "*" ->
      update lexbuf lexbuf';
      MUL
  | "/" ->
      update lexbuf lexbuf';
      DIV
  | "^" ->
      update lexbuf lexbuf';
      POW
  | "<" (blank | '\n') ->
      update lexbuf lexbuf';
      LT
  | "<=" ->
      update lexbuf lexbuf';
      LE
  | ">" ->
      update lexbuf lexbuf';
      GT
  | ">=" ->
      update lexbuf lexbuf';
      GE
  | '<' -> xml_tag_start stack lexbuf' lexbuf
  | eof ->
      update lexbuf lexbuf';
      EOF
  | _ ->
      update lexbuf lexbuf';
      Err.syntax (Pos.of_lexbuf lexbuf')
and xml_tag stack lexbuf' = lexer
  | blank ->
      xml_tag stack lexbuf' lexbuf
  | '\n' ->
      update lexbuf lexbuf';
      incr_line lexbuf';
      xml_tag stack lexbuf' lexbuf
  | "(*" ->
      update lexbuf lexbuf';
      code_comment lexbuf' lexbuf;
      xml_tag stack lexbuf' lexbuf
  | "/*" ->
      update lexbuf lexbuf';
      long_comment lexbuf' lexbuf;
      xml_tag stack lexbuf' lexbuf
  | "//" ->
      update lexbuf lexbuf';
      line_comment lexbuf' lexbuf;
      xml_tag stack lexbuf' lexbuf
  | ident -> lex_ident stack lexbuf' lexbuf
  | ':' -> COLON
  | '=' -> EQ
  | '(' -> lex_par_open stack lexbuf' lexbuf
  | '{' -> lex_curly_open stack lexbuf' lexbuf
  | '"' ->
      update lexbuf lexbuf';
      STRING (double_quoted_string (Buffer.create string_sz) lexbuf' lexbuf)
  | '\'' ->
      update lexbuf lexbuf';
      simple_quoted_string (Buffer.create string_sz) lexbuf' lexbuf
  | "..." ->
      update lexbuf lexbuf';
      DOTDOTDOT
  | "/>" ->
      pop stack lexbuf' "<" ">";
      XML_ELEMENT_END
  | '>' ->
      pop stack lexbuf' "<" ">";
      push stack lexbuf' "" (Lex xml_contents);
      XML_TAG_END
and xml_tag_start stack lexbuf' lexbuf =
  update lexbuf lexbuf';
  push stack lexbuf' "<" (Lex xml_tag);
  XML_TAG_START
and xml_contents stack lexbuf' = lexer
  | "</" ->
      pop stack lexbuf' "" "</";
      push stack lexbuf' "</" (Lex xml_tag_end);
      XML_ELEMENT_END
  | "<" -> xml_tag_start stack lexbuf' lexbuf
  | "{" -> lex_curly_open stack lexbuf' lexbuf
  | [^ "<" "{"]+ ->
      STRING (Ulexing.utf8_lexeme lexbuf)
and xml_tag_end stack lexbuf' = lexer
  | blank ->
      xml_tag stack lexbuf' lexbuf
  | '\n' ->
      update lexbuf lexbuf';
      incr_line lexbuf';
      xml_tag_end stack lexbuf' lexbuf
  | "(*" ->
      update lexbuf lexbuf';
      code_comment lexbuf' lexbuf;
      xml_tag_end stack lexbuf' lexbuf
  | "/*" ->
      update lexbuf lexbuf';
      long_comment lexbuf' lexbuf;
      xml_tag_end stack lexbuf' lexbuf
  | "//" ->
      update lexbuf lexbuf';
      line_comment lexbuf' lexbuf;
      xml_tag_end stack lexbuf' lexbuf
  | ':' -> COLON
  | '>' ->
      pop stack lexbuf' "</" ">";
      XML_TAG_END
  | ident -> lex_ident stack lexbuf' lexbuf
and lex_par_open stack lexbuf' lexbuf =
  update lexbuf lexbuf';
  push stack lexbuf' "(" (Lex main);
  PAR_OPEN
and lex_curly_open stack lexbuf' lexbuf =
  update lexbuf lexbuf';
  push stack lexbuf' "{" (Lex main);
  CURLY_OPEN
